/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */

#include "Player.hpp"
#include <limits>
#include <algorithm>

Player::Player(Block* b) :
	block(b),
	turnHoriz(0),
	turnVert(0),
	forwardPressed(false),
	backwardPressed(false),
	leftPressed(false),
	rightPressed(false),
	turned(false),
	moveSpeed(0.0625),
	playerRadius(0.125)
{
	//hrm, assumes playerRadius < 1.
	for(int i = 0; i < 3; ++i) {
		partialBlocks[i] = 0;
		position[i] = 0;
		for(int j = 0; j < 3; ++j)
			viewMatrix[i][j] = (i==j)?1:0;
	}
	
}
Player::Player(Block* b, Vector3d& p) :
	block(b),
	turnHoriz(0),
	turnVert(0),
	forwardPressed(false),
	backwardPressed(false),
	leftPressed(false),
	rightPressed(false),
	turned(false),
	moveSpeed(0.0625),
	playerRadius(0.125)
{
	for(int i = 0; i < 3; ++i) {
		partialBlocks[i] = 0;//FIXME
		position[i] = p[i];
		for(int j = 0; j < 3; ++j)
			viewMatrix[i][j] = (i==j)?1:0;
	}
}
		

std::pair<double,double> Player::look(DataStructure::Vertex_iterator begin, DataStructure::Vertex_iterator end)
{
	double max = -std::numeric_limits<double>::infinity();
	double min = std::numeric_limits<double>::infinity();
	
	for(DataStructure::Vertex_iterator i = begin; i != end; ++i) {
		Vector3d direction;
		direction[0] = i->point().x();
		direction[1] = i->point().y();
		direction[2] = 1.0;
		normalise(direction);
		direct(direction);
		i->info() = block->trace(direction, position);
		min = std::min(i->info(),min);
		max = std::max(i->info(),max);
	}
	return std::make_pair(min,max);
}

double Player::look(double x, double y)
{
	Vector3d direction;
	direction[0] = x;
	direction[1] = y;
	direction[2] = 1.0;
	normalise(direction);
	direct(direction);
	return block->trace(direction, position);
}

void Player::direct(Vector3d& direction)
{
	Vector3d temp;
	for(int i = 0; i < 3; ++i) temp[i] = direction[i];
	for(int i = 0; i < 3; ++i) {
		direction[i] = 0;
		for(int j = 0; j < 3; ++j)
			direction[i] += viewMatrix[i][j]*temp[j];
	}
}

void Player::turn(double horiz, double vert)
{
	turned = true;
	double sensitivity = 0.00390625;
	turnHoriz += horiz*sensitivity;
	turnVert += vert*sensitivity;
}

void Player::turn() {
	if(turned) {
		/*
		horizMatrix =
		| cos(phi)  0  sin(phi)|
		|   0       1     0    |
		|-sin(phi)  0  cos(phi)|
		
		vertMatrix =
		| 1       0           0    |
		| 0  cos(theta) -sin(theta)|
		| 0  sin(theta)  cos(theta)|
		
		viewMatrix = horizMatrix*vertMatrix
		*/
		double sinHoriz = sin(turnHoriz);
		double sinVert = sin(turnVert);
		double cosHoriz = cos(turnHoriz);
		double cosVert = cos(turnVert);
		
		viewMatrix[0][0] = cosHoriz;
		viewMatrix[0][1] = sinHoriz*sinVert;
		viewMatrix[0][2] = sinHoriz*cosVert;
		viewMatrix[1][0] = 0;
		viewMatrix[1][1] = cosVert;
		viewMatrix[1][2] = -sinVert;
		viewMatrix[2][0] = -sinHoriz;
		viewMatrix[2][1] = sinVert*cosHoriz;
		viewMatrix[2][2] = cosHoriz*cosVert;

		turned=false;
	}
}
//void turn(d h,d v) { tH += h; tV += v; }

bool Player::freeToPass(unsigned dimension, bool increasing)
{
	bool free = block->neighbour(dimension, !increasing) != 0;
	//Hmm, block is <0, not >0...
	for(int i = 0; i < 3; ++i) {
		if(partialBlocks[i] != 0) {
			free = free && (partialBlocks[i]->neighbour(dimension, !increasing) != 0);
		}
	}
	return free;
}

void Player::move(Vector3d& v)
{
	//Eww
	
	bool increasing=false;
	double distanceCanMove = std::numeric_limits<double>::infinity();
	int dimension = -1;
	HitNature nature = Invalid;//TODO: put in an enum
	
	for (int i = 0; i < 3; ++i) {
		if(v[i] == 0) continue;
		
		bool direction= (v[i] > 0);
		double p = position[i] * (direction?1:-1);
		double vel = v[i] * (direction?1:-1);
		HitNature thisNature;
		double thisDistance;
		if(p > (block->dim(i)-playerRadius) || ((p == (block->dim(i)-playerRadius)) && partialBlocks[i]!=0)) {
			//TODO: think about this test a bit more, get a cleaner definition
			//At the moment if we're on the border we're entering if we're outside,
			//crossing a face if we're inside (by testing whether we've entered)
			thisNature = Face;
			thisDistance = (block->dim(i)-p)/vel;
		} else if (p <= (-block->dim(i)+playerRadius) && (partialBlocks[i] != 0)) {
			thisNature = ExitBoundary;
			thisDistance = (-block->dim(i)+playerRadius-p)/vel;
		} else {
			thisNature = EnterBoundary;
			thisDistance = (block->dim(i)-playerRadius-p)/vel;
		}
		if(thisDistance < distanceCanMove) {
			distanceCanMove = thisDistance;
			increasing = direction;
			nature = thisNature;
			dimension = i;
		}
	}
	if(distanceCanMove < 1) {
		//Move up to whatever feature is stopping us
		position[0] += v[0]*distanceCanMove;
		position[1] += v[1]*distanceCanMove;
		position[2] += v[2]*distanceCanMove;
		
		switch(nature) {
			case EnterBoundary:
				//Check for wall
				if(freeToPass(dimension, increasing)) {
					//Add a neighbour if we can go there
					partialBlocks[dimension] = block->neighbour(dimension, !increasing);
				} else {
					v[dimension] = 0;
					//Hit a wall, motion in that direction stops
				}
				break;
			case Face:
				//The block in front of us and the block we're in swap places
				position[dimension] *= -partialBlocks[dimension]->dim(dimension)/block->dim(dimension);
				std::swap(block, partialBlocks[dimension]);
				//Our other neighbours shift along one
				for(int i = 0; i < 3; i++) {
					if(i != dimension && partialBlocks[i] != 0) {
						partialBlocks[i] = partialBlocks[i]->neighbour(dimension, !increasing);
					}
				}
				//Instead of being at the end of the block we're at the start
				
				break;
			case ExitBoundary:
				//We're no longer adjacent to this neighbour
				partialBlocks[dimension] = 0;
				break;
			default:
				std::cout << "what" << std::endl;
		}
		//Decrease distance to move by distance moved
		v[0] *= (1-distanceCanMove);
		v[1] *= (1-distanceCanMove);
		v[2] *= (1-distanceCanMove);

		//Move the rest of the way, now that we've dealt with the obstacle
		move(v);
	} else {
		//Move the whole way, there are no obstacles
		position[0] += v[0];
		position[1] += v[1];
		position[2] += v[2];
	}
}

void Player::update()
{
	//do player turning for the frame
	turn();

	//Get player input velocity
	Vector3d v;
	v[2] = (forwardPressed?1:0) - (backwardPressed?1:0);
	v[0] = (rightPressed?1:0) - (leftPressed?1:0);
	v[1] = 0;
	if((forwardPressed ^ backwardPressed) || (leftPressed ^ rightPressed)) {
		direct(v);//TODO: change to just horizontal when we get gravity
		normalise(v, moveSpeed);
	}
	
	move(v);
	//TODO: adjust v for accel (inc. G), remember c too.
	
	//find wall collision, distance.
	//if distance > 1
		//move
	//else if wall is wall
		//move(v*distance) //"*"? "/"?
		//adjust v (will be ours later)
		//recurse
	//else
		//move(v*distance)
		//move(wallside,-2)
		//recurse
	//(steal mostly from Block.cpp)
	
	//if(v[0] == 0 && v[1] ==0 && v[2] == 0) return;
	
}
